import logging

from .config_handler import ConfigHandler

def get_logger(name):

	config = ConfigHandler()
	logging.basicConfig(level=config.logging_mode)

	logger = logging.getLogger(name)
	#command_handler = logging.StreamHandler()
	#command_format = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
	#command_handler.setFormatter(command_format)
	#logger.addHandler(command_handler)
	return logger



from .config_handler import ConfigHandler
from .logger import get_logger
from .keyword_query import KeywordQuery
from .similarity import Similarity
from .graph import Graph
from .memory import memory_size,memory_percent
from .timestr import timestr
from .tokenizer import Tokenizer
from .next_path import next_path,last_path
from .tf_iaf import calculate_tf,calculate_iaf,calculate_inverse_frequency
from .truncate import truncate
from .mongo_query_builder import MongoQueryBuilder
from .document_traverser import DocumentTraverser
from .utils import (
  AttributeTreeGenerator,
  DictToSTupleFormatter,
  load_datasets_ids_from_json,
  load_schema_graph_from_json,
  modify_retrieved_documents,
)
from .ordinal import ordinal
from .printmd import printmd
from .shift_tab import shift_tab


import matplotlib
import matplotlib.pyplot as plt
import numpy as np


def grouped_bar_plot():
    #observations
    n=4
    #labels /colors
    m=4

    labels = [f'L{i}' for i in range(n)]
    observation_label = [f'O{i}' for i in range(m)]

    data = np.random.randint(10,40, size=(m,n))

    width  = 1/(m)
    inner_margin = 0.10
    outer_margin = 0.3

    x = np.arange(0,len(labels),1+outer_margin)# the label locations

    _fig, ax = plt.subplots()
    for i in range(m):
        ax.bar(x+( (1-m)/2 + i)*width, data[i], width*(1-2*inner_margin), label=observation_label[i])

    ax.set_xticks(x)
    ax.set_xticklabels(labels)

    ax.set_ylabel('Scores')
    ax.set_title('Scores by group and gender')

    ax.legend()
    plt.savefig('../../data/plots/teste.pdf')



import shelve
from random import sample

from sereia.index.babel_hash import BabelHash
from sereia.utils import (
    calculate_tf,
    calculate_iaf,
    calculate_inverse_frequency,
)

class ValueIndex():

    def __init__(self,**kwargs):
        self._dict = {}
        self.persistant_filename = kwargs.get('persistant_filename',None)
        self.update_cache = kwargs.get('update_cache',True)

    def __contains__(self, keyword):
        if keyword in self._dict:
            return True

        if self.persistant_filename:
            with shelve.open(self.persistant_filename,flag='r') as storage:
                if keyword in storage:
                    return True

        return False

    def _set_underlying_item(self, keyword, value):
        self._dict[keyword]=value

    def _get_underlying_item(self, keyword):
        item = None
        if keyword in self._dict:
            # print('Word in memory')
            item = self._dict[keyword]
        elif self.persistant_filename:
            # print('Word in disk')
            with shelve.open(self.persistant_filename,flag='r') as storage:
                item = storage.get(keyword,None)
                if self.update_cache:
                    self._set_underlying_item(keyword,item)
        if item is None:
            item = (0, BabelHash())
            # raise KeyError('The keyword {keyword} is not in the ValueIndex.')
        return item

    def __getitem__(self,keyword):
        _inverse_frequency, value = self._get_underlying_item(keyword)
        return value

    def __setitem__(self,keyword,value):
        raise Exception('Invalid operation. ValueIndex items shoud be set using the add_mapping method.')

    def __iter__(self):
        #This method only iterates over the keywords in memory
        yield from self._dict.keys()

    def keys(self):
        yield from self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self.__getitem__(key)

    def get_mappings(self,keyword,table,attribute):
        return self[keyword][table][attribute]

    def add_mapping(self,keyword,table,attribute,ctid):
        # This method does not change the persistant_filename
        self._dict.setdefault( keyword, (0, BabelHash() ) )
        self[keyword].setdefault(table , BabelHash() )
        self[keyword][table].setdefault( attribute , [] ).append(ctid)

    def get_inverse_frequency(self,keyword):
        inverse_frequency, _value = self._get_underlying_item(keyword)
        return inverse_frequency

    def set_inverse_frequency(self,keyword,inverse_frequency):
        # This method does not change the persistant_filename
        _old_inverse_frequency,old_value = self._dict[keyword]
        self._dict[keyword]= (inverse_frequency,old_value)

    def get_frequency(self,keyword,table,attribute):
        # print(keyword, table, attribute)
        # print('Babel Hash\n\n')
        # print(BabelHash.babel)
        # print('Keyword\n\n')
        # print(self[keyword])
        # print('Table\n\n')
        # print(self[keyword][table])
        # print('Attribute\n\n')
        # print(self[keyword][table][attribute])
        return len(self[keyword][table][attribute])

    def frequencies(self):
        for word in self:
            inverse_frequency = self.get_inverse_frequency(word)
            for table in self[word]:
                for attribute, ctids in self[word][table].items():
                    frequency = len(ctids)
                    yield table,attribute,frequency,inverse_frequency

    def get_iaf(self,weight_scheme,keyword):
        return calculate_iaf(weight_scheme, inverse_frequency = self.get_inverse_frequency(keyword))
    
    def get_tf(self,weight_scheme,keyword,table,attribute,max_frequency=None):
        return calculate_tf(weight_scheme,self.get_frequency(keyword,table,attribute), max_frequency = max_frequency)

    def __repr__(self):
        return f'<ValueIndex {repr(self._dict)}>'

    def persist_to_file(self, file_directory, filename):
        filepath = file_directory + filename
        with shelve.open(filepath) as storage:
            storage['__babel__']=BabelHash.babel
            for key,underlying_value in self._dict.items():
                storage[key]=underlying_value

    def load_from_file(self,persistant_filename,**kwargs):
        '''
        The load_method specifies the set of keywords to me loaded, which can be:
        - keywords: A list or set of keywords to be loaded. If the list is empty,
        no keyword is loaded in this method.
        - sample: A sample of keywords from shelve is loaded. The number of keywords
        in the sample is defined by sample_size, which is 15 by default.
        - all_keywords: All the keywords from shelve are loaded. Beware that this
         load methods might be expensive.
        '''
        self.persistant_filename = persistant_filename
        load_method = kwargs.get('load_method','keywords')
        keywords = kwargs.get('keywords',[])
        sample_size = kwargs.get('sample_size',15)

        with shelve.open(self.persistant_filename,flag='r') as storage:

            if load_method != 'keywords':
                keywords = storage.keys()

                if load_method == 'sample':
                    keywords = sample(keywords, k = sample_size)

            if '__babel__' in storage:
                BabelHash.babel.update(storage['__babel__'])

            for keyword in keywords:
                if keyword == '__babel__':
                    continue
                try:
                    self._set_underlying_item(keyword,storage[keyword])
                except KeyError:
                    self._set_underlying_item(keyword,None)
                    print(f'Keyword {keyword} not present in persistant_filename')
                    continue



from pathlib import Path
import pickle

from sereia.utils import Graph


class SchemaGraph(Graph):
    def __init__(self, graph_dict=None,edges_info = None):
        super().__init__(graph_dict,edges_info)

    def add_fk_constraint(self,constraint,cardinality,table,foreign_table,attribute_mappings):
        self.add_vertex(table)
        self.add_vertex(foreign_table)

        edge_info = self._edges_info.setdefault( 
            (table,foreign_table),
            {}
        ) 
        edge_info[constraint] = (cardinality,attribute_mappings)
        self.add_edge(table, foreign_table, edge_info)
        # print(edge_info)
        # print(f'Table: {table} | Foreign table: {foreign_table}')

    def add_root(self, table):
        self.add_vertex(table)

    def __repr__(self):
        if len(self)==0:
            return 'EmptyGraph'
        print_string = ['\t'*level+direction+vertex for direction,level,vertex in self.leveled_dfs_iter()]
        return '\n'.join(print_string)

    def persist_to_file(self, filepath, filename):
        Path(filepath).mkdir(parents=True, exist_ok=True)
        data = (self._graph_dict,self._edges_info)
        with open(filepath + filename, mode='wb') as f:
            pickle.dump(data,f)

    def load_from_file(self,filename):
        with open(filename,mode='rb') as f:
            data = pickle.load(f)
        self._graph_dict,self._edges_info = data


import csv
import json



BASE_EXPERIMENT_FILE = '../results/cns_from_golden_qms/{}-standard.json'

class CSVReport(object):

    def __init__(self, dataset_name, results):
        self.dataset_name = dataset_name
        self.results = results

    def _load_results(self):
        dataset_results_filename = BASE_EXPERIMENT_FILE.format(
            self.dataset_name,
        )

        with open(dataset_results_filename) as f:
            self.results = json.load(f)

    def generate(self):

        if 'retrieval_score' not in self.results['evaluation']:
            return

        result_per_query = self.results['results']
        evaluation_results = self.results['evaluation']['retrieval_score']

        overall_results = {}

        for item in result_per_query:
            keyword_query = item['keyword_query']
            query_generation_time = item['query_generation_time']
            query_execution_time = item['query_execution_time']
            query_number = item['query_number']

            overall_results[keyword_query] = {
                'query_gen_time': query_generation_time,
                'query_exec_time': query_execution_time,
                'query_number': query_number
            }
        
        for item in evaluation_results:
            keyword_query = item['keyword_query']
            precision = item['precision']
            recall = item['recall']
            retrieved_count = item['num_documents_retrieved']
            expected_count = item['num_documents_expected']

            overall_results[keyword_query]['precision'] = precision
            overall_results[keyword_query]['recall'] = recall
            overall_results[keyword_query]['retrieved_count'] = retrieved_count
            overall_results[keyword_query]['expected_count'] = expected_count
        
        print('Generating CSV report')
        from pprint import pprint as pp
        pp(overall_results)

        report_filepath = '../query_results.csv'

        with open(report_filepath, 'w') as f:
            field_names = [
                'Keyword Query',
                'Query Number',
                'Precision',
                'Recall',
                'Gen Time (s)',
                'Exec Time (s)',
                '# Doc. Retrieved',
                '# Doc. Expected',
            ]

            writer = csv.DictWriter(f, fieldnames=field_names, quoting=csv.QUOTE_ALL)

            writer.writeheader()

            for keyword_query in overall_results:
                writer.writerow(
                    {
                        'Keyword Query': keyword_query,
                        'Query Number': overall_results[keyword_query]['query_number'] + 1,
                        'Precision': overall_results[keyword_query]['precision'],
                        'Recall': overall_results[keyword_query]['recall'],
                        'Gen Time (s)': overall_results[keyword_query]['query_gen_time'],
                        'Exec Time (s)': overall_results[keyword_query]['query_exec_time'],
                        '# Doc. Retrieved': overall_results[keyword_query]['retrieved_count'],
                        '# Doc. Expected': overall_results[keyword_query]['expected_count']
                    }
                )
            





from math import log

'''
Weight Schemes recommended by Baeza-Yates

| Scheme | TF Weight                | IDF Weight        |
|--------|--------------------------|-------------------|
| 0      | raw frequency            | inverse frequency |
| 1      | log normalization        | unary             |
| 2      | log normalization        | inverse frequency |
| 3      | double normalization 0.5 | inverse frequency |

'''

def calculate_tf(weight_scheme,frequency,max_frequency=None):
    if weight_scheme == 0:
        # Raw Frequency
        return frequency
    if weight_scheme == 3:
        # Double Normalization 0.5
        return 0.5 + 0.5*frequency/max_frequency
    if weight_scheme in [1,2]:
        # Log Normalization
        return 1+log(frequency)
    raise ValueError(f'Invalid weight_scheme({weight_scheme})')

def calculate_inverse_frequency(num_total_attributes,num_attributes_with_this_word):
    return log(num_total_attributes/num_attributes_with_this_word)

def calculate_iaf(weight_scheme,inverse_frequency=None):
    if weight_scheme == 1:
        # Unary
        return 1
    if weight_scheme in [0,2,3]:
        # Inverse Frequency
        return inverse_frequency
    raise ValueError(f'Invalid weight_scheme({weight_scheme})')
    
    


import json
from collections import Counter
from queue import deque

from sereia.keyword_match import KeywordMatch
from sereia.utils import (
    Graph,
    MongoQueryBuilder,
    AttributeTreeGenerator,
    load_datasets_ids_from_json,
)


class CandidateNetwork(Graph):
    def __init__(self, graph_dict=None):
        self.score = None
        self.__root = None

        super().__init__(graph_dict)

        if len(self)>0:
            self.set_root()

    def get_root(self):
        return self.__root

    def set_root(self,vertex=None):
        if len(self) == 0:
            return None

        if vertex is not None:
            keyword_match,_ = vertex
            if not keyword_match.is_free():
                self.__root = vertex
                return vertex
            else:
                return None
        else:         
            for candidate in self.vertices():
                keyword_match,_ = candidate
                if not keyword_match.is_free():
                    self.__root = candidate
                    return candidate
        
        print('The root of a Candidate Network cannot be a Keyword-Free Match.')
        print(self)
        raise ValueError('The root of a Candidate Network cannot be a Keyword-Free Match.')
    
    def reset_root(self, to_value=True):
        if to_value:
            for leaf in self.leaves():
                keyword_match, _ = leaf
                if keyword_match.has_value_mappings():
                    self.__root = leaf

    def get_starting_vertex(self):
        if len(self)==0:
            return None

        if self.__root is None:
            self.set_root()
        return self.__root

    def get_starting_vertex(self, rev):
        if len(self) == 0:
            return None

        if self.__root is None:
            self.set_root()
        return self.__root
        
    def add_vertex(self, vertex):
        results = super().add_vertex(vertex)
        if self.__root is None:
            self.set_root(vertex)
        return results

    def add_keyword_match(self, heyword_match, **kwargs):
        alias = kwargs.get('alias', 't{}'.format(self.__len__()+1))
        vertex = (heyword_match, alias)
        return self.add_vertex(vertex)

    def add_adjacent_keyword_match(self,parent_vertex,keyword_match,edge_direction='>',**kwargs):
        child_vertex = self.add_keyword_match(keyword_match,**kwargs)
        self.add_edge(parent_vertex,child_vertex,edge_direction=edge_direction)

    def keyword_matches(self):
        for keyword_match,_ in self.vertices():
            yield keyword_match

    def non_free_keyword_matches(self):
        for keyword_match,_ in self.vertices():
            if not keyword_match.is_free():
                yield keyword_match

    def num_free_keyword_matches(self):
        i=0
        for keyword_match,_ in self.vertices():
            if keyword_match.is_free():
                i+=1
        return i

    def reciprocal_neighbors(self,schema_graph,vertex):
        keyword_match,_ = vertex
        for neighbor_vertex in self.incoming_neighbors(vertex):
            neighbor_km,_ = neighbor_vertex

            edge_info = schema_graph.get_edge_info(
                neighbor_km.table,
                keyword_match.table
            )            
            for _,(cardinality,_) in edge_info.items():
                if cardinality=='1:1':
                    yield neighbor_vertex
        

    def is_sound(self,schema_graph):
        if len(self) < 3:
            return True

        def has_duplicate_tables(neighbors):
            if len(neighbors)>=2:
                tables = set()
                for neighbor_km,_ in neighbors:
                    if neighbor_km.table not in tables:
                        tables.add(neighbor_km.table)
                    else:
                        return True
            return False

        for vertex in self.vertices():
            outgoing_neighbors = list(self.outgoing_neighbors(vertex))
            if has_duplicate_tables(outgoing_neighbors):
                return False

            reciprocal_neighbors = list(self.reciprocal_neighbors(schema_graph,vertex))
            if has_duplicate_tables(reciprocal_neighbors):
                return False
        return True


    def remove_vertex(self,vertex):
        _,incoming_neighbors = self._graph_dict[vertex]
        for neighbor in incoming_neighbors:
            self._graph_dict[neighbor][0].remove(vertex)
        self._graph_dict.pop(vertex)

    def is_total(self,query_match):
        return set(self.non_free_keyword_matches())==set(query_match)

    def contains_keyword_free_match_leaf(self):
        for vertex in self.leaves():
            keyword_match,_ = vertex
            if keyword_match.is_free():
                return True
        return False

    def minimal_cover(self,query_match):
        return self.is_total(query_match) and not self.contains_keyword_free_match_leaf()

    def unaliased_edges(self):
        for (keyword_match,_),(neighbor_keyword_match,_) in self.edges():
            yield (keyword_match,neighbor_keyword_match)

    def calculate_score(self, query_match):
        self.score = query_match.total_score/len(self)

    def get_qm_score(self):
        return self.score*len(self)

    def __eq__(self, other):

        if not isinstance(other,CandidateNetwork):
            return False
        
        self_root_km,_  = self.__root
        other_root_km,_= other.get_root()

        other_hash = None

        if self_root_km==other_root_km:
            other_hash = hash(other)
        else:            
            for keyword_match,alias in other.vertices():
                if self_root_km == keyword_match:
                    root = (keyword_match,alias)
                    other_hash = other.hash_from_root(root)
        
        if other_hash is None:
            return False

        return hash(self)==other_hash

    def __hash__(self):
        if len(self)==0:
            return hash(None)
        if self.__root is None:
            self.set_root()
        return self.hash_from_root(self.__root)   


    def hash_from_root(self,root):
        hashable = []

        level = 0       
        visited = set()

        queue = deque()
        queue.append( (level,root) )

        while queue:
            level,vertex = queue.popleft()
            keyword_match,alias = vertex
            children = Counter()
            visited.add(alias)
            
            for adj_vertex in self.neighbors(vertex):
                adj_km,adj_alias = adj_vertex
                if adj_alias not in visited:
                    queue.append( (level+1,adj_vertex) )
                    children[adj_km]+=1
                
            if len(hashable)<level+1:
                hashable.append(set())
            
            hashable[level].add( (keyword_match,frozenset(children.items())) )

        hashcode = hash(tuple(frozenset(items) for items in hashable))
        return hashcode   

    def __repr__(self):
        if len(self)==0:
            return 'EmptyCN'
        print_string = ['\t'*level+direction+str(vertex[0])  for direction,level,vertex in self.leveled_dfs_iter()]
        return '\n'.join(print_string)

    def to_json_serializable(self):
        return [{'keyword_match':keyword_match.to_json_serializable(),
            'alias':alias,
            'outgoing_neighbors':[alias for (km,alias) in outgoing_neighbors],
            'incoming_neighbors':[alias for (km,alias) in incoming_neighbors]}
            for (keyword_match,alias),(outgoing_neighbors,incoming_neighbors) in self._graph_dict.items()]

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_json_serializable(json_serializable_cn):
        alias_hash ={}
        edges=[]
        for vertex in json_serializable_cn:
            keyword_match = KeywordMatch.from_json_serializable(vertex['keyword_match'])
            alias_hash[vertex['alias']]=keyword_match

            for outgoing_neighbor in vertex['outgoing_neighbors']:
                edges.append( (vertex['alias'],outgoing_neighbor) )

        candidate_network = CandidateNetwork()
        for alias,keyword_match in alias_hash.items():
            candidate_network.add_vertex( (keyword_match,alias) )
        for alias1, alias2 in edges:
            vertex1 = (alias_hash[alias1],alias1)
            vertex2 = (alias_hash[alias2],alias2)
            candidate_network.add_edge(vertex1,vertex2)

        return candidate_network

    @staticmethod
    def from_json(json_cn):
        return CandidateNetwork.from_json_serializable(json.loads(json_cn))

    def get_mongo_query_from_cn(self, database_name, schema_graph):
        selected_attributes = set()
        selected_tables = []

        next_step_lookup_field = None
        last_added_collection = None

        datasets_collection_ids_dict = load_datasets_ids_from_json()

        query = MongoQueryBuilder(database_name)
        self.reset_root(to_value=True)

        fields_to_project = []

        for prev_vertex, direction, vertex in self.dfs_pair_iter(root_predecessor=True, rev=False, start_vertex=self.__root):
            keyword_match, alias = vertex
            project_fields = []
            value_fields = []

            tree_generator = AttributeTreeGenerator()

            for type_km, _, attribute, keywords in keyword_match.mappings():
                selected_attributes.add(f'{alias}.{attribute}')
                if type_km == 'v':
                    for keyword in keywords:
                        if isinstance(keyword, str):
                            sql_keyword = keyword.replace('\'', '\'\'')
                            value_fields.append((attribute, sql_keyword))
                        elif isinstance(keyword, frozenset):
                            # print('Found froze: {}'.format(keyword))
                            value_fields.append((attribute, keyword))
                            # exit(0)
                    continue

            if prev_vertex is None:
                query.set_starting_collection(keyword_match.table)
                selected_tables.append(f'{keyword_match.table} {alias}')
                edge_info = schema_graph.get_edge_info(keyword_match.table,
                                                       keyword_match.table)
                if edge_info:
                    for constraint in edge_info:
                        _, attribute_mappings = edge_info[constraint]
                        if type(attribute_mappings) != list:
                            l = []
                            l.append(attribute_mappings)
                            attribute_mappings = l
                        for constraint_column, foreign_column in attribute_mappings:
                            project_fields.append(constraint_column)
                for value_field in value_fields:
                    tree_generator.add_attribute(value_field[0], value_field[1])
                tree_generator.generate()
                query.new_match_v3(keyword_match.table, tree_generator.generate())
            else:
                _, prev_alias = prev_vertex
                if direction == '>':
                    constraint_keyword_match, constraint_alias = prev_vertex
                    foreign_keyword_match, foreign_alias = vertex
                else:
                    constraint_keyword_match, constraint_alias = vertex
                    foreign_keyword_match, foreign_alias = prev_vertex

                edge_info = schema_graph.get_edge_info(constraint_keyword_match.table,
                                                       foreign_keyword_match.table)

                for constraint in edge_info:
                    _, attribute_mappings = edge_info[constraint]

                    if type(attribute_mappings) != list:
                        l = []
                        l.append(attribute_mappings)
                        attribute_mappings = l

                    if last_added_collection == keyword_match.table:
                        continue

                    if not next_step_lookup_field:
                        constraint_column, foreign_column = attribute_mappings[0]
                        query.lookup(constraint_column, foreign_column,
                                     keyword_match.table, keyword_match.table)
                    else:
                        constraint_column, foreign_column = attribute_mappings[0]
                        query.lookup(f'{next_step_lookup_field}.{constraint_column}',
                                     foreign_column, keyword_match.table, keyword_match.table)

                if last_added_collection != keyword_match.table:
                    query.unwind(keyword_match.table, insert=True)
                    if value_fields:
                        for attribute, sql_keyword in value_fields:
                            attribute_name = keyword_match.table + '.' + attribute
                            tree_generator.add_attribute(attribute, sql_keyword)
                        query.new_match_v3(keyword_match.table, tree_generator.generate())
                        next_step_lookup_field = keyword_match.table
                    else:
                        next_step_lookup_field = keyword_match.table
                else:
                    for attribute, sql_keyword in value_fields:
                        tree_generator.add_attribute(attribute, sql_keyword)
                    query.new_match_v3(keyword_match.table, tree_generator.generate())

            if next_step_lookup_field:
                if keyword_match.table in datasets_collection_ids_dict[database_name]:
                    for attribute in datasets_collection_ids_dict[database_name][keyword_match.table]:
                        fields_to_project.append(
                            next_step_lookup_field + '.' + attribute,
                        )
            else:
                fields_to_project.append(datasets_collection_ids_dict[database_name][keyword_match.table][0])
            
            last_added_collection = keyword_match.table

        query.project(
            fields_to_project,
        )

        return query



class KeywordQuery(object):

  __slots__ = ['keyword_query', 'parsed_keyword_query', 'keywords', 'length']

  def __init__(self, parsed_keyword_query, keyword_query):
    self.parsed_keyword_query = parsed_keyword_query
    self.keyword_query = keyword_query

  
  def get_value(self):
    return self.keyword_query
  
  def get_parsed_value(self):
    return self.parsed_keyword_query
  
  # def get_compound_segments(self):


  def get_keywords(self):

    if hasattr(self, 'keywords'):
      return self.keywords

    keywords = []

    for element in self.parsed_keyword_query:
      if isinstance(element, list):
        for item in element:
          keywords.append(item.replace('\"', ''))
        continue
      
      keywords.append(element.replace('\"', ''))
    
    self.keywords = keywords
    self.length = len(keywords)

    return keywords



import json
from pathlib import Path
from os import listdir
from os.path import isfile, join
from timeit import default_timer as timer

from sereia.candidate_network import CandidateNetworkHandler
from sereia.database import MongoHandler
from sereia.evaluation import EvaluationHandler
from sereia.index import IndexHandler
from sereia.keyword_match import KeywordMatchHandler
from sereia.query_match import QueryMatchHandler
from sereia.utils import ConfigHandler, KeywordQuery, Similarity, Tokenizer
from sereia.utils.result import SereiaResult


class Sereia():
    
    def __init__(self, database_name, database_credentials, config_directory):
        self.config = ConfigHandler(
            config_directory,
        )
        self.queryset_configs = self.config.get_queryset_configs()
        self.max_num_query_matches = 100
        self.max_qm_size = 3
        self.max_cjn_size = 3
        self.topk_cns = 10
        self.topk_cns_per_qm = 1
        self.topk_cns_per_qm_list = [1]
        self.approaches = ['standard']
        self.assume_golden_qms = False
        self.input_desired_cn = False
        self.datasets_configurations = {}
        self.querysets_configurations = {}
        self.current_database = database_name

        #dont forget to choose the Database according to the config file
        self.database_handler = MongoHandler(
            self.current_database,
            database_credentials,
            self.config,
        )
        self.index_handler = IndexHandler(
            self.current_database,
            self.database_handler,
            self.config,
        )
        self.tokenizer = Tokenizer(tokenize_method = 'simple')
        self.similarity = Similarity(self.index_handler.schema_index)
        self.keyword_match_handler = KeywordMatchHandler(self.similarity)
        self.query_match_handler = QueryMatchHandler()
        self.evaluation_handler = EvaluationHandler(self.current_database, self.config)
        self.candidate_network_handler = CandidateNetworkHandler(self.database_handler)
        self.evaluation_handler.load_golden_standards()

    def list_datasets(self):
        datasets = self.database_handler.get_databases()
        print(datasets)
        # datasets_folder = DefaultConfiguration.DATASETS_FOLDER
        # datasets_configurations = [(f.replace('.json', ''), join(datasets_folder, f)) for f in listdir(datasets_folder) if isfile(join(datasets_folder, f))]
        # for configuration in datasets_configurations:
        #     self.datasets_configurations[configuration[0]] = configuration[1] 

        # print(sorted(list(self.datasets_configurations.keys())))
    
    def list_querysets(self):
        querysets_folder = './querysets/' #DefaultConfiguration.QUERYSETS_FOLDER
        querysets_configurations = [(f.replace('.json', ''), join(querysets_folder, f)) for f in listdir(querysets_folder) if isfile(join(querysets_folder, f))]
        for configuration in querysets_configurations:
            self.querysets_configurations[configuration[0]] = configuration[1] 

        print(sorted(list(self.querysets_configurations.keys())))

    def use_database(self, database_name):
        self.current_database = database_name
        self.database_handler.set_database_name(self.current_database)

    def use_queryset(self, queryset_name):
        self.current_queryset = queryset_name
        with open(self.config.queryset_filepath, mode='r') as f:
            self._queryset = json.load(f)

    def create_indexes(self):
        self.index_handler.create_indexes()

    def load_indexes(self):
        self.index_handler.load_indexes()

    def run_queryset(self, **kwargs):
        results_filename = kwargs.get('results_filename',None)
        export_results = kwargs.get('export_results',False)
        approach = kwargs.get('approach','standard')
        preprocessed_results = kwargs.get('preprocessed_results',{})

        results =[]

        keywords_to_load = {keyword for item in self.get_queryset() for keyword in set(self.tokenizer.keywords(item['keyword_query']))}

        self.index_handler.load_indexes(keywords = keywords_to_load)

        for item in self.get_queryset():
            keyword_query = item['keyword_query']

            if keyword_query in preprocessed_results:
                result = preprocessed_results[keyword_query]
            else:
                result = self.keyword_search(keyword_query,**kwargs)
            results.append(result)

        data = {
            "database":self.config.connection['database'],
            "queryset":self.config.queryset_filepath,
            "results":results,
        }

        if export_results:
            if results_filename is None:
                results_filename = 'output.json'#next_path(f'{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json')

            with open(results_filename,mode='w') as f:
                print(f'Writing results in {results_filename}')
                json.dump(data,f, indent = 4)

        return data

    def keyword_search(self, keyword_query=None, **kwargs):
        max_qm_size = kwargs.get('max_qm_size', self.max_qm_size)
        max_num_query_matches = kwargs.get('max_num_query_matches', self.max_num_query_matches)
        max_cjn_size = kwargs.get('max_cjn_size', self.max_cjn_size)
        topk_cns = kwargs.get('topk_cns', self.topk_cns)
        topk_cns_per_qm = kwargs.get('topk_cns_per_qm', self.topk_cns_per_qm)
        weight_scheme = kwargs.get('weight_scheme', 0)

        repeat = kwargs.get('repeat', 1)
        assume_golden_qms = kwargs.get('assume_golden_qms', False)
        
        input_desired_cn = kwargs.get('input_desired_cn', False)
        skip_cn_generations = kwargs.get('skip_cn_generations', False)
        show_kms_in_result = kwargs.get('show_kms_in_result', True)
        use_result_class = kwargs.get('use_result_class', True)

        weight_scheme = kwargs.get('weight_scheme', 3)

        if 'weight_scheme' in kwargs:
            del kwargs['weight_scheme']

        generated_query = None
        retrieved_documents = []

        elapsed_time = {
            'km':[],
            'skm':[],
            'vkm':[],
            'qm':[],
            'cn':[],
            'total':[],
        }

        if keyword_query is None:
            print(f'Please input a keyword query or choose one of the queries below:')
            for i,item in enumerate(self._queryset):
                keyword_query = item['keyword_query']
                print(f'{i+1:02d} - {keyword_query}')
            return None

        if isinstance(keyword_query, int):
            keyword_query=self._queryset[keyword_query - 1]['keyword_query']

        keywords = self.tokenizer.extract_keywords(keyword_query)
        keyword_query = KeywordQuery(keywords, keyword_query)

        for _ in range(repeat):
            
            if not assume_golden_qms:
                start_skm_time = timer()
                
                sk_matches = self.keyword_match_handler.schema_keyword_match_generator(
                    keyword_query,
                    self.index_handler.schema_index,
                )
                
                start_vkm_time = timer()
                vk_matches = self.keyword_match_handler.value_keyword_match_generator(
                    keyword_query,
                    self.index_handler.value_index,
                )
                # vk_matches = self.keyword_match_handler.filter_kwmatches_by_compound_keywords(vk_matches,compound_keywords)

                kw_matches = sk_matches + vk_matches
                start_qm_time = timer()

                query_matches = self.query_match_handler.generate_query_matches(
                    keyword_query.get_keywords(),
                    kw_matches,
                )
            else:
                start_skm_time = timer()
                start_vkm_time = timer()
                start_qm_time = timer()
                kw_matches = []
                query_matches = self.evaluation_handler.golden_standards[keyword_query]['query_matches']

            
            ranked_query_matches = self.query_match_handler.rank_query_matches(query_matches,
                self.index_handler.value_index,
                self.index_handler.schema_index,
                self.similarity,
                weight_scheme,
            )

            ranked_query_matches = ranked_query_matches[:max_num_query_matches]

            start_cn_time = timer()

            if input_desired_cn:
                desired_cn = self.evaluation_handler.golden_standards[keyword_query]['candidate_networks'][0]
                kwargs['desired_cn'] = desired_cn
            else:
                kwargs['desired_cn'] = None

            if not skip_cn_generations:
                ranked_cns = self.candidate_network_handler.generate_cns(
                    self.index_handler.schema_index,
                    self.index_handler.schema_graph,
                    ranked_query_matches,
                    keywords,
                    weight_scheme,
                        **kwargs,
                )

                if len(ranked_cns) > 0:
                    top_cn = ranked_cns[0]
                    base_collection, generated_query = top_cn.get_mongo_query_from_cn(
                        self.current_database,
                        self.index_handler.schema_graph,
                    ).build()
                    query_result = self.database_handler.execute(base_collection, generated_query)
                    retrieved_documents.extend(query_result)
            else:
                ranked_cns=[]

            end_cn_time = timer()

            elapsed_time['skm'].append(start_vkm_time - start_skm_time)
            elapsed_time['vkm'].append(start_qm_time - start_vkm_time)
            elapsed_time['km'].append(start_qm_time - start_skm_time)
            elapsed_time['qm'].append(start_cn_time - start_qm_time)
            elapsed_time['cn'].append(end_cn_time - start_cn_time)
            elapsed_time['total'].append(end_cn_time - start_skm_time)

        aggregated_elapsed_time = {phase:min(times) for phase,times in elapsed_time.items()}

        result = {
            # 'keyword_query': keyword_query,
            'keywords': list(keywords),
            # 'compound_keywords':list(compound_keywords),
            'query_matches':      [query_match.to_json_serializable()
                                  for query_match in ranked_query_matches],
            'candidate_networks': [candidate_network.to_json_serializable()
                                  for candidate_network in ranked_cns],
            'elapsed_time':       aggregated_elapsed_time,
            'num_keyword_matches':len(kw_matches),
            #consider len of unranked query matches
            'num_query_matches':  len(query_matches),
            'num_candidate_networks':  len(ranked_cns),
            'generated_query': generated_query,
        }

        if show_kms_in_result:
            result['value_keyword_matches'] = [vkm.to_json_serializable() for vkm in vk_matches]
            result['schema_keyword_matches']= [skm.to_json_serializable() for skm in sk_matches]

        return SereiaResult(self.current_database, self.index_handler, self.database_handler, result)



from pathlib import Path
import pickle

import pymongo

from sereia.utils import (
    ConfigHandler,
    DocumentTraverser,
    Tokenizer,
)


class MongoIter:
    def __init__(self, database_table_columns, database_name, database_client,
                 database_cursor_batch_size=10000, **kwargs):
        self.database_name = database_name
        self.database_client = database_client
        self.database_table_columns = database_table_columns
        self.table_hash = self._get_indexable_schema_elements()
        self.limit_per_collection = kwargs.get('limit_per_table', 0)
        self.tokenizer = kwargs.get('tokenizer', Tokenizer())
        self.database_cursor_batch_size = database_cursor_batch_size

    def _schema_element_validator(self,table,column):
        return True

    def _get_indexable_schema_elements(self):

        table_hash = {}
        for table, column in self.database_table_columns:
            table_hash.setdefault(table, []).append(column)

        return table_hash

    def __iter__(self):
        collection_count = 0
        for collection, attributes in self.table_hash.items():
            collection_count += 1

            indexable_attributes = [
                 attr for attr in attributes if self._schema_element_validator(collection, attr)]

            if len(indexable_attributes) == 0:
                continue

            print(f"{collection_count} of {len(self.table_hash.keys())} collections indexed")

            database = self.database_client[self.database_name]
            collection_document_count = database[collection].count_documents({})
            document_count = 0

            traverser = DocumentTraverser()

            for document in database[collection].find(
                projection=attributes,
                batch_size=self.database_cursor_batch_size,
                show_record_id=True,
            ):

                document_record_id = document['$recordId']
                del document['$recordId']

                document_count += 1

                if (document_count % self.database_cursor_batch_size) == 0:
                    print(f"Indexed {document_count} of {collection_document_count} for collection {collection}")

                traverser.traverse(document)
                indexable_content = traverser.get_indexable_content()

                for column, content in indexable_content:
                    content = str(content)

                    tokens = self.tokenizer.tokenize(content)

                    for token in tokens:
                        yield collection, document_record_id, column, token

                traverser.cleanup()



from csv import DictWriter, QUOTE_ALL
import json
from sys import maxsize as SYS_MAXSIZE
import os

INPUT_DATA_FILES = [
    #('results/1cns_per_qm/yelp-standard.json', 'yelp',),
    ('../../../results/cns_from_golden_qms/imdb-standard.json', 'imdb.json',),
    ('../../../results/cns_from_golden_qms/nba-standard.json', 'nba.json',),
    ('../../../results/cns_from_golden_qms/dblp-standard.json', 'dblp.json',),
    ('../../../results/cns_from_golden_qms/twitter-standard.json', 'twitter.json',),
    ('../../../results/cns_from_golden_qms/yelp_expanded-standard.json', 'yelp_expanded.json',),
]

print(os.getcwd())

for file, dataset_name in INPUT_DATA_FILES:

    print(f'Formatting evaluation results from {dataset_name} dataset')

    with open(file) as f:
        evaluation_results = json.load(f)

    header = [
        'query_number',
        'vkm_count',
        'skm_count',
        'qm_count',
        'cn_count',
        # 'max(cn_count)',
        # 'min(cn_count)',
    ]

    # for i in range(29):
    #     header.append(str(i + 1))

    with open(f'parsed_results_{dataset_name}.csv', 'w') as f:
        writer = DictWriter(f, fieldnames=header)
        writer.writeheader()

        results = evaluation_results['results']
        for n, result in enumerate(results):
            data = {}
            data['query_number'] = f'Query {n}'

            if 'vkm_count' in result:
                data['vkm_count'] = result['vkm_count']
            else:
                data['vkm_count'] = 0
            if 'skm_count' in result:
                data['skm_count'] = result['skm_count']
            else:
                data['skm_count'] = 0

            data['qm_count'] = result['qm_count']
            data['cn_count'] = result['cn_count']

            max_cn_count = 0
            min_cn_count = SYS_MAXSIZE

            # for qm in result['statistics']:
            #     current_cn_count = result['statistics'][qm]['generated_cns_qty']
            #     if current_cn_count > max_cn_count:
            #         max_cn_count = current_cn_count

            #     if current_cn_count < min_cn_count:
            #         min_cn_count = current_cn_count

            # data['max(cn_count)'] = max_cn_count
            # data['min(cn_count)'] = min_cn_count
            writer.writerow(data)

    header = [
        'query_number',
        'correct_qm_rank',
        'correct_cn_rank',
    ]

    with open(f'formatted_evaluation_{dataset_name}.csv', 'w') as f:
        writer = DictWriter(f, fieldnames=header)
        writer.writeheader()

        qm_relevant_positions = evaluation_results['evaluation']['query_matches']['relevant_positions']
        cn_relevant_positions = evaluation_results['evaluation']['candidate_networks']['relevant_positions']
        qm_relevant_positions_len = len(qm_relevant_positions)
        cn_relevant_positions_len = len(cn_relevant_positions)

        if qm_relevant_positions_len != cn_relevant_positions_len:
            print(f'Different list lengths for QMs ({qm_relevant_positions_len}) and CNs ({cn_relevant_positions_len})!')
            exit(-1)

        number_of_queries = qm_relevant_positions_len
        for n in range(number_of_queries):
            data = {}
            data['query_number'] = f'Query {n + 1}'
            data['correct_qm_rank'] = qm_relevant_positions[n]
            data['correct_cn_rank'] = cn_relevant_positions[n]
            writer.writerow(data)

    # header = [
    #     'qm_evaluation_metric',
    #     'qm_result',
    #     'cn_evaluation_metric',
    #     'cn_result',
    # ]

    # with open(f'formatted_metric_results_{dataset_name}.csv', 'w') as f:
    #     writer = DictWriter(f, fieldnames=header)
    #     writer.writeheader()

    #     lathe_steps = [
    #         'query_matches',
    #         'candidate_networks'
    #     ]

    #     evaluation_metric_keys = evaluation_results['evaluation']['query_matches'].keys()

    #     for key in evaluation_metric_keys:
    #         if key == 'relevant_positions':
    #             continue

    #         data = {}
    #         data['qm_evaluation_metric'] = key
    #         data['qm_result'] = evaluation_results['evaluation']['query_matches'][key]
    #         data['cn_evaluation_metric'] = key
    #         data['cn_result'] = evaluation_results['evaluation']['candidate_networks'][key]
    #         writer.writerow(data)

    # header = [
    #     'query_number',
    #     'vkm_phase',
    #     'skm_phase',
    #     'qm_phase',
    #     'cn_phase',
    #     'total_time'
    # ]

    # with open(f'formatted_per_phase_time_{dataset_name}.csv', 'w') as f:
    #     writer = DictWriter(f, fieldnames=header, quoting=QUOTE_ALL)
    #     writer.writeheader()

    #     results = evaluation_results['results']
    #     for n, result in enumerate(results):
    #         data = {}
    #         data['query_number'] = f'Query Number {n + 1}'
    #         data['vkm_phase'] = '{:.3f}'.format(result['elapsed_time']['vkm']).replace('.', ',')
    #         data['skm_phase'] = '{:.3f}'.format(result['elapsed_time']['skm']).replace('.', ',')
    #         data['qm_phase'] = '{:.3f}'.format(result['elapsed_time']['qm']).replace('.', ',')
    #         data['cn_phase'] = '{:.3f}'.format(result['elapsed_time']['cn']).replace('.', ',')
    #         data['cn_phase'] = '{:.3f}'.format(result['elapsed_time']['cn']).replace('.', ',')
    #         data['total_time'] = '{:.3f}'.format(result['elapsed_time']['total']).replace('.', ',')
    #         writer.writerow(data)

    # results = evaluation_results['results']
    # vkm_phase = []
    # skm_phase = []
    # qm_phase = []
    # cn_phase = []
    # total_time = []
    # for n, result in enumerate(results):
    #     print(result['elapsed_time']['vkm'], result['elapsed_time']['skm'], result['elapsed_time']['qm'], result['elapsed_time']['cn'], result['elapsed_time']['total'])
    #     vkm_phase.append(result['elapsed_time']['vkm'])
    #     skm_phase.append(result['elapsed_time']['skm'])
    #     qm_phase.append(result['elapsed_time']['qm'])
    #     cn_phase.append(result['elapsed_time']['cn'])
    #     total_time.append(result['elapsed_time']['total'])

    # print('Avg VKM time: {}'.format(sum(vkm_phase)/len(vkm_phase)))
    # print('Avg SKM time: {}'.format(sum(skm_phase)/len(skm_phase)))
    # print('Avg QM time: {}'.format(sum(qm_phase)/len(qm_phase)))
    # print('Avg CN time: {}'.format(sum(cn_phase)/len(cn_phase)))
    # print('Avg total time: {}'.format(sum(total_time)/len(total_time)))
